<a href={{ pathto(master_doc) }}><img class="logo" src={{ pathto("_static/nngt_logo.svg",1) }}></a>
<p class="toctitle">Table of content</p>
<ul class="globaltoc">
    {{ toctree(maxdepth=-1, collapse=True,includehidden=True) }}
</ul>
<br />
