{% extends "!layout.html" %}
{% block extrahead %}
    {{ super() }}
    <meta name="keywords" content="nngt, neuron, network, graph, topology, neuronal, python, nest" />
    <meta name="description" content="Documentation for the python module NNGT, aimed at describing neuronal networks and interfacing them with simulators." />
{% endblock %}
